#! /usr/bin/python
# Copyright (c) 2018 Arista Networks, Inc.  All rights reserved.
# Arista Networks, Inc. Confidential and Proprietary.

import argparse
import os
import subprocess
import sys

import Logging
import PyClient
import Tracing

# Eng defined values ===
# Public name of advisory
secAdvName = "SecurityAdvisory0034"
KMOD = 'xt_TCPMSS'
KMOD_ORIG = 'xt_TCPMSS.orig.ko'
KMOD_FIX = 'xt_TCPMSS.fixed.ko'
SWI_VER = 'SWI_VERSION=4.20.1FX-Virtual-Router'
KERNEL_VER = '3.18.28.Ar-6792879.4201FXVirtualRouter'
# End Eng defined values ===
hotfixFilePath = "/etc/{secAdvName}/".format( secAdvName=secAdvName )

Logging.logD( id="SECURITY_HOTFIX_ERR",
              severity=Logging.logError,
              format="Issue with Hotfix %s: %s",
              explanation="Log issue with hotfix that should be followed up on",
              recommendedAction=Logging.CALL_SUPPORT_IF_PERSISTS )
Logging.logD( id="SECURITY_HOTFIX_INFO",
              severity=Logging.logInfo,
              format="Hotfix %s: %s",
              explanation="Log info about hotfix that does not " +\
              " require further action",
              recommendedAction=Logging.NO_ACTION_REQUIRED )
traceHandle = Tracing.Handle( "SecAdvHotfix" )
traceInfo = traceHandle.trace2

def recordMsg( msg, isError ):
   """
   Double emits message, both to logging and stderr.
   """
   # pylint: disable-msg=E0602
   logHandle = None
   if isError:
      logHandle = SECURITY_HOTFIX_ERR
   else:
      logHandle = SECURITY_HOTFIX_INFO
   Logging.log( logHandle, secAdvName, msg )
   traceInfo( msg )
   # pylint: enable-msg=E0602

def die( err ):
   """
   Record the given error message and exit with error code 1.
   """
   recordMsg( err, isError=True )
   sys.exit( 1 )

def fileContains( filename, content ):
   """
   Return true if the filename can be read and contains the given content.
   """
   try:
      with open( filename ) as f:
         return content in f.read()
   except IOError:
      return False

def runCmd( cmd ):
   """
   Execute the given command array and record an error message and exit if it
   failed.
   """
   try:
      p = subprocess.Popen( cmd,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT )
      ret = p.wait()
   except OSError as e:
      die( 'Could not run "{}": {}'.format( ' '.join( cmd ), e.strerror ) )
   if ret:
      die( 'Command "{}" failed with {}: {}'.format( ' '.join( cmd ),
                                                     ret, p.stdout.read() ) )

def saveMssConfig( tun ):
   """
   Retrieve the TCP MSS values for every configured tunnel and reset to their
   default value.
   """
   res = []
   for t in tun:
      if tun[ t ].maxMss:
         res.append( ( t, tun[ t ].maxMss ) )
         tun[ t ].maxMss = 0
   return res

def restoreMssConfig( tun, prevMss, force=False ):
   """
   Restore the TCP MSS values for every configured tunnel. When forced, ignore
   exceptions to try hard to restore all the original values.
   """
   if not tun:
      return
   for t, mss in prevMss:
      try:
         tun[ t ].maxMss = mss
      except Exception: # pylint: disable-msg=W0703
         recordMsg( 'Could not restore MSS value {} for {}'.format( mss, t ),
                    isError=True )
         if not force:
            raise

def main():
   parser = argparse.ArgumentParser(
            description="Install fix for Arista " + secAdvName )
   parser.add_argument( "--install", action="store_true",
                        help="Install hotfix" )
   parser.add_argument( "--uninstall", action="store_true",
                        help="Uninstall hotfix" )
   args = parser.parse_args()

   # make sure we have exactly one of {args.install,args.uninstall}
   if args.install == args.uninstall:
      parser.print_help()
      sys.exit( 1 )
   if not ( fileContains( '/etc/swi-version', SWI_VER ) and
            os.uname()[ 2 ] == KERNEL_VER ):
      die( 'The fix is not compatible with this version of EOS.' )

   if args.install:
      newModule = os.path.join( hotfixFilePath, KMOD_FIX )
      action = 'installed'
   else:
      newModule = os.path.join( hotfixFilePath, KMOD_ORIG )
      action = 'uninstalled'

   if fileContains( '/proc/modules', KMOD ):
      tun = None
      mss = []
      try:
         pc = PyClient.PyClient( "ar", "Sysdb" )
         root = pc.agentRoot()
         tun = root.entity.get( 'interface/config/tunnel/intf' )
         mss = saveMssConfig( tun )
         if mss:
            recordMsg( 'Saved MSS values: {}'.format( mss ), isError=False )
            # wait for TunnelIntfAclPlugin to remove all the TCPMSS rules
            runCmd( [ 'wfw', 'Acl' ] )
         runCmd( [ 'rmmod', KMOD ] )
         runCmd( [ 'insmod', newModule ] )
         restoreMssConfig( tun, mss )
         # original config restored, don't do it a second time in the finally
         mss = []
      except ( PyClient.RpcError, ValueError, KeyError, AttributeError ):
         die( 'Could not access, save or restore tunnel MSS ceiling values.' )
      finally:
         # if something failed, try at least to restore the original config
         restoreMssConfig( tun, mss, force=True )
   else:
      runCmd( [ 'insmod', newModule ] )
   recordMsg( "Successfully {}".format( action ), isError=False )

if __name__ == "__main__":
   main()
